#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <clutter/clutter.h>
#include <tidy/tidy.h>

enum
{
  COLUMN_FOO,
  COLUMN_BAR,
  COLUMN_BAZ,

  N_COLUMNS
};

static void
append_clicked_cb (TidyButton *button,
                   ClutterModel *model)
{
  ClutterModelIter *iter;
  gboolean foo_value = FALSE;
  gint bar_value = 0;
  gchar *baz_value = NULL;

  iter = clutter_model_get_last_iter (model);
  if (!iter)
    return;

  clutter_model_iter_get (iter, COLUMN_BAR, &bar_value, -1);

  bar_value += 1;
  foo_value = (bar_value % 2 == 0 ? TRUE : FALSE);
  baz_value = g_strdup_printf ("<b>My string %d</b>", bar_value);

  clutter_model_append (model,
                        COLUMN_FOO, foo_value,
                        COLUMN_BAR, bar_value,
                        COLUMN_BAZ, baz_value,
                        -1);

  g_free (baz_value);
}

static void
remove_clicked_cb (TidyButton   *button,
                   ClutterModel *model)
{
  clutter_model_remove (model, 0);
}

static void
change_clicked_cb (TidyButton   *button,
                   ClutterModel *model)
{
  ClutterModelIter *iter;
  gboolean foo_value = FALSE;
  gint bar_value = 0;
  gchar *baz_value = NULL;

  iter = clutter_model_get_last_iter (model);
  if (!iter)
    return;

  clutter_model_iter_get (iter,
                          COLUMN_FOO, &foo_value,
                          COLUMN_BAR, &bar_value,
                          COLUMN_BAZ, &baz_value,
                          -1);

  foo_value = !foo_value; 
  bar_value += 100;
  baz_value = g_strdup_printf ("<i>Changed string %d</i>", bar_value);

  clutter_model_iter_set (iter,
                          COLUMN_FOO, foo_value,
                          COLUMN_BAR, bar_value,
                          COLUMN_BAZ, baz_value,
                          -1);
}

typedef struct {
  ClutterModel *model;
  gint n_rows;
  gint cur_row;
} PopulateIdle;

static gboolean
populate_model_idle (gpointer data)
{
  PopulateIdle *closure = data;
  gchar *str;

  str = g_strdup_printf ("<b>String %d</b>", closure->cur_row + 1);

  clutter_model_append (closure->model,
                        COLUMN_FOO, (closure->cur_row % 2),
                        COLUMN_BAR, closure->cur_row,
                        COLUMN_BAZ, str,
                        -1);
  g_free (str);

  closure->cur_row += 1;

  if (closure->cur_row == closure->n_rows)
    return FALSE;
  else
    return TRUE;
}

static void
populate_model_clean (gpointer data)
{
  PopulateIdle *closure = data;

  g_object_unref (closure->model);

  g_slice_free (PopulateIdle, closure);
}

static ClutterModel *
create_model (void)
{
  PopulateIdle *closure;
  ClutterModel *model;

  closure = g_slice_new0 (PopulateIdle);

  model = clutter_list_model_new (N_COLUMNS,
                                  G_TYPE_BOOLEAN, "Boolean Foo",
                                  G_TYPE_INT,     "Integer Bar",
                                  G_TYPE_STRING,  "String Baz");
  closure->model = g_object_ref (model);
  closure->n_rows = 15;
  closure->cur_row = 0;

  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE + 30,
                   populate_model_idle, closure,
                   populate_model_clean);

  return model;
}

static void
on_list_view_row_clicked (TidyListView     *list_view,
                          ClutterModel     *model,
                          ClutterModelIter *iter)
{
  g_print ("Row : %d clicked\n", clutter_model_iter_get_row (iter));
}

static ClutterActor *
create_view (ClutterModel *model)
{
  ClutterActor *view;
  TidyListView *list_view;
  guint w, h;
  gint i;

  view = tidy_list_view_new (model);
  list_view = TIDY_LIST_VIEW (view);
  tidy_list_view_set_show_headers (list_view, TRUE);
  g_signal_connect (list_view, "row-clicked",
                    G_CALLBACK (on_list_view_row_clicked),
                    NULL);

  for (i = 0; i < N_COLUMNS; i++)
    {
      tidy_list_view_add_column (list_view,
                                 tidy_list_column_default_new (list_view, i));
    }

  w = 700;
  h = 400;

  clutter_actor_set_size (view, w, h);
  clutter_actor_set_position (view, (800 - w) / 2, (600 - h) / 2);

  return view;
}

static ClutterActor *
create_buttons (ClutterModel *model)
{
  ClutterActor *retval;
  ClutterActor *button;
  guint w, h;

  retval = clutter_group_new ();

  w = 700;
  h = 100;

  clutter_actor_set_size (retval, w, h);
  clutter_actor_set_position (retval, (800 - w) / 2, (600 - h));

  button = tidy_button_new_with_label ("Append a row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_actor_set_position (button, 0, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (append_clicked_cb),
                    model);

  button = tidy_button_new_with_label ("Remove a row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_set_position (button, 200, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (remove_clicked_cb),
                    model);

  button = tidy_button_new_with_label ("Change the last row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_set_position (button, 400, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (change_clicked_cb),
                    model);

  return retval;
}

int
main (int argc, char *argv[])
{
  ClutterActor *stage, *view, *box;
  ClutterModel *model;
  ClutterColor stage_color = { 0xff, 0xff, 0xff, 0xff };

  clutter_init (&argc, &argv);

  stage = clutter_stage_get_default ();
  clutter_stage_set_color (CLUTTER_STAGE (stage), &stage_color);
  clutter_actor_set_size (stage, 800, 600);
  clutter_actor_show (stage);

  model = create_model ();
  view = create_view (model);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), view);
  clutter_actor_show (view);
  
  box = create_buttons (model);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), box);
  clutter_actor_show (box);

  clutter_main ();

  g_object_unref (model);

  return EXIT_SUCCESS;
}
