#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <clutter/clutter.h>
#include <tidy/tidy.h>

enum
{
  COLUMN_FOO,
  COLUMN_BAR,
  COLUMN_BAZ,

  N_COLUMNS
};

#define RECT_W 300
#define RECT_H 300
#define RECT_N 200
#define RECT_GAP 50

static void
viewport_x_origin_notify_cb (TidyViewport *viewport,
                             GParamSpec *args1,
                             ClutterActor *group)
{
  GList *children, *c;
  gint origin_x, width;
  
  tidy_viewport_get_origin (viewport, &origin_x, NULL, NULL);
  width = clutter_actor_get_width (
            clutter_actor_get_parent (CLUTTER_ACTOR (viewport)));
  
  children = clutter_container_get_children (CLUTTER_CONTAINER (group));
  for (c = children; c; c = c->next)
    {
      gint x;
      gdouble pos;
      ClutterActor *actor;
      
      actor = (ClutterActor *)c->data;
      
      /* Get actor position with respect to viewport origin */
      x = clutter_actor_get_x (actor) - origin_x;
      pos = (((gdouble)x / (gdouble)(width-RECT_W)) - 0.5) * 2.0;
      
      /* Apply a function that transforms the actor depending on its 
       * viewport position.
       */
      pos = CLAMP(pos * 3.0, -0.5, 0.5);
      clutter_actor_set_rotation (actor, CLUTTER_Y_AXIS, -pos * 120.0,
                                  RECT_W/2, 0, RECT_H/2);
      clutter_actor_set_depth (actor, -ABS(pos) * RECT_W);
    }
  g_list_free (children);
}

static void
add_rects (ClutterActor *stage, ClutterActor *group)
{
  gint i;
  
  ClutterColor colour = { 0x72, 0x9f, 0xcf, 0xff };

  for (i = 0; i < RECT_N; i++)
    {
      ClutterActor *rect = clutter_rectangle_new_with_color (&colour);
      
      clutter_container_add_actor (CLUTTER_CONTAINER (group), rect);

      clutter_actor_set_size (rect, RECT_W, RECT_H);
      clutter_actor_set_position (rect, i * RECT_GAP,
                                  CLUTTER_STAGE_HEIGHT ()/2 - RECT_H/2);

      colour.red = g_random_int_range (0, 255);
      colour.green = g_random_int_range (0, 255);
    }
}

static void
append_clicked_cb (TidyButton *button,
                   ClutterModel *model)
{
  ClutterModelIter *iter;
  gboolean foo_value = FALSE;
  gint bar_value = 0;
  gchar *baz_value = NULL;

  iter = clutter_model_get_last_iter (model);
  if (!iter)
    return;

  clutter_model_iter_get (iter, COLUMN_BAR, &bar_value, -1);

  bar_value += 1;
  foo_value = (bar_value % 2 == 0 ? TRUE : FALSE);
  baz_value = g_strdup_printf ("<b>My string %d</b>", bar_value);

  clutter_model_append (model,
                        COLUMN_FOO, foo_value,
                        COLUMN_BAR, bar_value,
                        COLUMN_BAZ, baz_value,
                        -1);

  g_free (baz_value);
}

static void
remove_clicked_cb (TidyButton   *button,
                   ClutterModel *model)
{
  clutter_model_remove (model, 0);
}

static void
change_clicked_cb (TidyButton   *button,
                   ClutterModel *model)
{
  ClutterModelIter *iter;
  gboolean foo_value = FALSE;
  gint bar_value = 0;
  gchar *baz_value = NULL;

  iter = clutter_model_get_last_iter (model);
  if (!iter)
    return;

  clutter_model_iter_get (iter,
                          COLUMN_FOO, &foo_value,
                          COLUMN_BAR, &bar_value,
                          COLUMN_BAZ, &baz_value,
                          -1);

  foo_value = !foo_value; 
  bar_value += 100;
  baz_value = g_strdup_printf ("<i>Changed string %d</i>", bar_value);

  clutter_model_iter_set (iter,
                          COLUMN_FOO, foo_value,
                          COLUMN_BAR, bar_value,
                          COLUMN_BAZ, baz_value,
                          -1);
}

static ClutterModel *
create_model (void)
{
  ClutterModel *model;
  gint i;

  model = clutter_list_model_new (N_COLUMNS,
                                  G_TYPE_BOOLEAN, "Boolean Foo",
                                  G_TYPE_INT,     "Integer Bar",
                                  G_TYPE_STRING,  "String Baz");

  for (i = 0; i < 10; i++)
    {
      gchar *str = g_strdup_printf ("<b>My string %d</b>", i);

      clutter_model_append (model,
                            COLUMN_FOO, (i % 2 == 0),
                            COLUMN_BAR, i,
                            COLUMN_BAZ, str,
                            -1);

      g_free (str);
    }

  return model;
}

static ClutterActor *
create_view (ClutterModel *model)
{
  ClutterActor *view, *scroll;
  TidyListView *list_view;
  guint w, h;

  view = tidy_list_view_new (model);
  list_view = TIDY_LIST_VIEW (view);

  tidy_list_view_add_column (list_view,
                             tidy_list_column_default_new (list_view, 0));

  scroll = tidy_finger_scroll_new (TIDY_FINGER_SCROLL_MODE_KINETIC);
  clutter_container_add_actor (CLUTTER_CONTAINER (scroll), view);
  clutter_actor_show (view);

  w = 700;
  h = 400;

  clutter_actor_set_size (scroll, w, h);
  clutter_actor_set_position (scroll, (800 - w) / 2, (600 - h) / 2);

  return scroll;
}

static ClutterActor *
create_buttons (ClutterModel *model)
{
  ClutterActor *retval;
  ClutterActor *button;
  guint w, h;

  retval = clutter_group_new ();

  w = 700;
  h = 100;

  clutter_actor_set_size (retval, w, h);
  clutter_actor_set_position (retval, (800 - w) / 2, (600 - h));

  button = tidy_button_new_with_label ("Append a row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_actor_set_position (button, 0, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (append_clicked_cb),
                    model);

  button = tidy_button_new_with_label ("Remove a row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_set_position (button, 200, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (remove_clicked_cb),
                    model);

  button = tidy_button_new_with_label ("Change the last row");
  tidy_actor_set_alignment (TIDY_ACTOR (button), 0.5, 0.5);
  clutter_container_add_actor (CLUTTER_CONTAINER (retval), button);
  clutter_actor_set_position (button, 400, 0);
  clutter_actor_set_size (button, 150, 75);
  clutter_actor_show (button);

  g_signal_connect (button, "clicked",
                    G_CALLBACK (change_clicked_cb),
                    model);

  return retval;
}

int
main (int argc, char *argv[])
{
  ClutterActor *stage, *view, *box, *viewport, *group, *finger_scroll,
               *notebook, *label;
  ClutterModel *model;
  ClutterColor stage_color = { 0xff, 0xff, 0xff, 0xff };

  clutter_init (&argc, &argv);

  stage = clutter_stage_get_default ();
  clutter_stage_set_color (CLUTTER_STAGE (stage), &stage_color);
  clutter_actor_set_size (stage, 800, 600);
  clutter_actor_show (stage);

  model = create_model ();
  view = create_view (model);
  
  box = create_buttons (model);
  
  viewport = tidy_viewport_new ();
  group = tidy_depth_group_new ();
  add_rects (stage, group);
  clutter_container_add_actor (CLUTTER_CONTAINER (viewport), group);
  g_signal_connect (viewport, "notify::x-origin",
                    G_CALLBACK (viewport_x_origin_notify_cb), group);

  finger_scroll = tidy_finger_scroll_new (TIDY_FINGER_SCROLL_MODE_KINETIC);
  clutter_container_add_actor (CLUTTER_CONTAINER (finger_scroll), viewport);
  
  notebook = tidy_notebook_new ();
  clutter_actor_set_size (notebook, 700, 500);
  clutter_actor_set_position (notebook, 50, 50);

  label = tidy_button_new_with_label ("Tab 1");
  tidy_notebook_insert_page (TIDY_NOTEBOOK (notebook), label, view, 0);
  clutter_actor_show (label);
  label = tidy_button_new_with_label ("Tab 2");
  tidy_notebook_insert_page (TIDY_NOTEBOOK (notebook), label, finger_scroll, 1);
  clutter_actor_show (label);
  
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), notebook);

  clutter_actor_show (view);
  clutter_actor_show (box);
  clutter_actor_show_all (group);
  clutter_actor_show (viewport);
  clutter_actor_show (finger_scroll);
  clutter_actor_show (notebook);
  
  g_object_notify (G_OBJECT (viewport), "x-origin");
  
  clutter_main ();

  g_object_unref (model);

  return EXIT_SUCCESS;
}
